"""
POWER BUSINESS SYSTEMS
Copyright 2010 Power Business Systems <jasper@powerbs.co.za>
All Rights Reserved.

Licensed under The Artistic License 2.0
See LICENSE for details or
http://www.perlfoundation.org/artistic_license_2_0
"""


"""
Core module for the bridge.
Waits for incoming connections (port number defined in clanrc.py)
and connects to hosts defined in conf/hosts_conf.py
"""

import os
import sys
import time
import socket
import re
from select import select
import thread
from clan_lib import clan_lib
from clan_lib import bridge_result 
from log import log
from copy import deepcopy
import traceback

__version__ = "0.35"

class clan_host (clan_lib, bridge_result) :
    pidfile = "clan.pid"
    cfgfile = "clanrc"
    running = False
    config = None
    hostname = ""
    server = None
    readers = []
    writers = []    
    errors = []
    clans = {}
    clans_lock = None
    daemons = {}
    send_queue = [] 
    event_listeners = []
    permissions = {}

    permissions_conf = {}
    daemons_conf = {}
    hosts_conf = {} 

    last_daemons_conf = ""
    last_permissions_conf = ""
    last_hosts_conf = ""

    def __init__ (self, hostname, host_ip, port):
        """
        hostname: The name of this bridge. Unrelated to OS hostname
        host_ip:  Ip address to listen on
        port:     Port number to listen on
        """
        self.clans_lock = thread.allocate_lock()
        self.time = deepcopy(time.time)
        self.sleep = deepcopy(time.sleep)
        self._read_rc()
        self._load_configs()
        # read rc now gets the hostname, remove it from __init__
        if hostname and hostname != "" :
            self.set_hostname (hostname)
        self._load_permissions()
        if port :
            self.start_listener(host_ip, port)
            log("Bridge started")
        self._start_daemons()

    def __del__ (self):
        try:
            self._close_all()
            log("Bridge stopped")
        except:
            pass

    def _close_all(self):
        """
        First stop all running daemons and then cycle through
        all active hosts and disconnect all of them.
        """
        self._stop_daemons()
        c = self.clans.copy()
        for s in c:
            self.remove_host(self.get_hostname_by_sock(s))
        self.server.close()

    def _load_permissions(self):
        """
        Turn the array groups into dict groups for faster lookups
        """
        p = self.permissions_conf.permissions
        for grp in p:
            self.permissions[grp] = {}
            for call in p[grp]:
                self.permissions[grp][call] = {} 

    def _load_configs(self):
        self._reload_configs()

    def _reload_configs(self):
        """
        Read config settings from conf/*_conf.py files but only if they have
        been change since the last read.
        """
        try:
            p_ts = os.path.getmtime("conf"+os.sep+"permissions_conf.py") 
            d_ts = os.path.getmtime("conf"+os.sep+"daemons_conf.py") 
            h_ts = os.path.getmtime("conf"+os.sep+"hosts_conf.py") 
            if p_ts != self.last_permissions_conf or d_ts != self.last_daemons_conf or h_ts != self.last_hosts_conf:
                try:
                    sys.modules.pop('conf.permissions_conf')
                    sys.modules.pop('conf.daemons_conf')
                    sys.modules.pop('conf.hosts_conf')
                    sys.modules.pop('conf')
                except:
                    pass
                m = __import__("conf.permissions_conf") 
                self.permissions_conf = m.permissions_conf 
                m = __import__("conf.daemons_conf") 
                self.daemons_conf = m.daemons_conf 
                m = __import__("conf.hosts_conf") 
                self.hosts_conf = m.hosts_conf
                self._load_permissions()
                self.last_permissions_conf = os.path.getmtime("conf"+os.sep+"permissions_conf.py")
                self.last_daemons_conf = os.path.getmtime("conf"+os.sep+"daemons_conf.py")
                self.last_hosts_conf = os.path.getmtime("conf"+os.sep+"hosts_conf.py")
        except:
            log("got error in _reload_configs")
            log(traceback.print_exc())        

    def _read_rc (self) :
        """
        Read the main config file, contains global config settings
        """
        if not self.cfgfile or not os.path.isfile(self.cfgfile + ".py") :
            log("Could not read/find rc file")
            return
        if sys.modules.has_key(self.cfgfile) :
            sys.modules.pop(self.cfgfile)
        self.config = __import__(self.cfgfile)
        self.hostname = self.config.hostname

 
    def _send_event(self, event_obj) :
        """
        Daemons can listen for certain events which get broadcast by _send_event
        Events can be new connections, lost connections etc
        Example usage could be:
        e = clan_event("host_connected", host=hostname)
        self._send_event(e)      
        where "host_connected" is the name of the event followed by **kwargs of
        event data        
        """
        for l in self.event_listeners:
            l.append(event_obj) 


    """
    def _mk_pidfile (self) :
        if not self.pidfile :
            return False
        try :
            fh = open(self.pidfile, "w+")
            fh.write(str(os.getpid()))
            fh.close()
        except :
            log("Could not create clan pid file: %s : %s" % (sys.exc_type, traceback.print_exc()))
            return False
        return True
    """

    def _start_daemons(self) :
        """
        Go through every entry in "conf/daemons_conf.py" and try to start each specified daemon 
        script (in daemons/)
        """
        d = self.daemons_conf.daemons
        for name in d:
            mod = d[name].get("module")
            cls = d[name].get("class")
            if not os.path.isfile("daemons/"+mod+".py") and not os.path.isfile("daemons/"+mod+".pyc") :
                log("Daemon: %s does not exist" % name)
                continue
            try :
                self._start_daemon(name,mod,cls)
            except :
                log("Daemon: %s - Could not daemonize" % name)
                log(traceback.print_exc())

    def _stop_daemons(self):
        d = self.daemons_conf.daemons
        for name in d:
            try :
                self._stop_daemon(name)
            except :
                log("Daemon: %s - Could not stop" % name)
                log(traceback.print_exc())

   
    def _start_daemon(self, name, mod="", cls="") :
        """
        Start a daemon script and attach it to self.daemons
        """
        d = self.daemons_conf.daemons
        if not d.get(name):
            log("Daemon: %s - No such daemon in config" % name)
            return False
        if mod == "" or cls == "":
            mod = d[name].get("module")
            cls = d[name].get("class") 
        if not mod or not cls:
            log("Daemon: %s - Could not start, no module or class" % name)
            return False
        try:
            # unloading the module might be bad, test
            try:
                sys.modules.pop(mod)
            except:
                pass
            c = __import__(mod)
            m = eval("c.%s(name,self)" % cls)
            self.daemons[name] = m
            m.start()
            log("Daemon: Started : %s" % name)
            return True 
        except:
            log("Daemon: '%s' broke" % name)
            log(traceback.print_exc())
            return False

    def _stop_daemon(self, name) :        
        if not self.daemons.has_key(name) :
            return False
        self.daemons[name].stop()
        self.daemons[name] = None
        log("Daemon: Stopped %s" % name)
        return True


    def connect (self, hostname, host_ip, port) :
        """
        Connect to another host and send authentication request
        Connection then gets stored in self.clans
        """
        self.remove_host(hostname)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(5)  
        try :
            sock.connect( (host_ip, int(port)) )
        except :
            return 
        log("connected to %s : %s " % (host_ip, port))
        if not sock :
            return 
        self.readers.append(sock)
        self.writers.append(sock)
        self.add_connection(sock, hostname, host_ip, port)
        self.send_register_req(hostname, sock)
        self.send_auth(hostname)

    def send_register_req (self, hostname, sock) :
        """
        Let the other host know who you are.
        Used to be necessary just before encrypting data so 
        the other host will know which key to use.
        Encryption will be added again soon.
        """
        self.remove_host(hostname, sock)
        req_str = self.create_rpc_str("reg", "", {"hn":self.hostname})
        sock = self.get_sock_by_hostname(hostname)
        self.queue_req(sock, req_str)

    def add_connection(self, sock, hostname, client_ip, client_port) :
        """
        Add a connection to self.clans
        The socket descriptor is the key in self.clans
        At this point the other host has not been authenticated yet
        Fires event 'host_connected'
        """
        self.clans_lock.acquire()
        self.clans[sock] = {
                "authenticated":False,
                "remote_authenticated":False,
                "hostname":hostname,
                "connected":True,
                "remote_auth_failed":False,
                "ip_address":client_ip,
                "port":client_port,
                "last_active":self.time(),
                "last_ping":self.time(),
                "send_queue":[],
                "read_lock":thread.allocate_lock(),
                "rfd":[]
            }
        
        self.clans_lock.release()
        self._read_rc() 
        e = clan_event("host_connected", host=hostname)
        self._send_event(e)       
        thread.start_new_thread(self.queue_handler, (sock,))      
 
    def send_auth (self, hostname):
        """
        Send authentication request to hostname
        """
        sock = self.get_sock_by_hostname(hostname)
        remote_key = self.get_remote_key(hostname) 
        auth_str = self.create_rpc_str("auth", "", {"host":self.hostname,"key":remote_key})
        self.queue_req(sock, auth_str)
        #sock.send(auth_str)

    def get_sock_by_rfd (self, rfd) :
        """
        Each rpc request gets allocated an rfd id(request for data)
        which is appended to the socket, which enables us to find
        the socket an rfd belongs to
        """
        self.clans_lock.acquire()
        for s in self.clans :
            for r in self.clans[s]["rfd"]:
                if r == rfd :
                    self.clans_lock.release()
                    return s
        self.clans_lock.release()
        return None     

    def get_sock_by_hostname (self, hostname):
        self.clans_lock.acquire()
        for s in self.clans :
            if self.clans[s]["hostname"] == hostname :
                self.clans_lock.release()
                return s
        self.clans_lock.release()
        return None     

    def get_hostname_by_sock (self, sock) :
        return self.clans[sock]["hostname"]

    def set_hostname (self, hostname) :
        self.hostname = hostname

    def start_listener (self, host_ip, port):
        """
        Open the listening port for incoming connections
        """
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind((host_ip,port))
        self.server.listen(50)

    def run (self) :
        """
        This is the main loop which is called after initializing the clan_host class
        Example:
        ch = clan_host(hostname, "", port) # run on all network interfaces
        ch.connect_clans()
        ch.run()
        """
        try:
            reconnect_secs = 10
            self.readers.append(self.server)
            self.writers.append(self.server)
            next_time = int(self.time()) + reconnect_secs
            self.running = True
            while self.running :
                try:
                    if not self.hostname:
                        self.sleep(3)
                        self._read_rc()
                        continue
                   
                    r, w, e = select(self.readers, self.writers, self.errors, 1)    
                 
                    for s_read in r :
                        if s_read == self.server :
                            client, address = self.server.accept()
                            log("Connection from %s" % address[0])
                            self.readers.append(client)
                            self.writers.append(client)
                            self.add_connection(client, "", address[0], address[1])
                            continue
                        thread.start_new_thread(self.handle_req, (s_read,))
                    self.sleep(0.05)
                    #self.check_queue()
                    if int(self.time()) >= next_time :
                        try:
                            self.connect_clans()
                        except:
                            log(sys.exc_value)
                            pass 
                        next_time = int(self.time()) + reconnect_secs 
                except:
                    log("-- Main loop error...")
                    log(sys.exc_value)
                    log("Shutting down main thread")
                    self.running = False
            self._close_all()
        except:
            log(traceback.print_exc())

    """
    def check_queue (self):
        " ""
        Check the internal queue for any unsent data
        "" "
        if len(self.send_queue) == 0 :
            return
        rdict = self.send_queue.pop(0)
        hostname = rdict["hostname"]
        req = rdict["req"]
        rfd = rdict["rfd"]
        sock = self.get_sock_by_hostname(hostname)
        if not sock :
            return
        log("---- check_queue: %s" % hostname)
        
        self.send_req(sock, req) 
        if rfd:
            self.remove_rfd(sock, rdict["rfd"])
    """

    def remove_host(self, hostname, exc_sock = None) :
        """
        Attempt to remove a connected host
        """
        if not hostname or hostname == "" :
            return
        rem_socks = []
        try :
            self.clans_lock.acquire()
            for s in self.clans :
                if exc_sock and exc_sock == s :
                    continue
                if self.clans[s]["hostname"] == hostname :
                    rem_socks.append(s)
            self.clans_lock.release()
            for rs in rem_socks :
                self.remove_connection(rs)

        except :
            log("ERROR in remove_host: %s" % traceback.print_exc())         

    def remove_connection (self, sock) :
        try :

            e = clan_event("host_removed", host=self.clans[sock]["hostname"])
            self._send_event(e)               
            self.clans_lock.acquire()
            self.clans.pop(sock)
            self.clans_lock.release()
        except :
            log("ERROR in remove_connection[clan]: %s" % traceback.print_exc())         
        try :
            self.readers.remove(sock)
            self.writers.remove(sock)
            sock.close()
            log("Removed connection")
        except :
            log("ERROR in remove_connection[sock]: %s" % traceback.print_exc())         

    def handle_req (self, sock) :
        """
        Reads a socket and attempts to parse the rpc request
        A defined handler will be called for each rpc request type
        """
        cnt = 300 # 30 second timeout
        l = self.clans[sock]["read_lock"] 
        while not l.acquire(False):
            cnt-= 1
            if cnt <= 0:
                log("handle_req: got a timeout")
                return
            if not self.clans.has_key(sock): # in case the connection was removed while waiting
                return 
            time.sleep(0.1) # wait and see if the lock gets released
        req_str = self.get_req(sock)
        l.release()
        if req_str == "" :
            self.remove_connection(sock)
            return
        br = bridge_result(req_str)
        data = br.get_row(0)
        req_type = data.get("_t")
        try :
            if not req_type :
                log("no request type was given")
                self.remove_connection(sock)
                return
            if req_type == "rpc" :
                return self.handle_rpc_req(sock, data) 
            if req_type == "rpc_reply" :
                return self.handle_rpc_reply(sock, data, br) 
            if req_type == "auth_reply" :
                return self.handle_auth_reply(sock, data)
            if req_type == "auth" :
                return self.handle_auth_req(sock, data)
            if req_type == "reg" :
                return self.handle_reg_req(sock, data)
            if req_type == "ping":
                pong = self.create_rpc_str("pong", "", {"time":self.time()})
                self.queue_req(sock, pong)

        except :
            log(traceback.print_exc())
            return False 
        return True 

    def handle_reg_req(self, sock, data) :
        """
        A remote host has sent register request to this host
        A register request is basically just to link a socket to a hostname 
        """
        hostname = data.get("hn")
        if not hostname :
            return
        self.remove_host(hostname, sock)
        self.clans[sock]["hostname"] = hostname

    def signal_auth(self,sock) :
        """
        Trigger the "host successfully authenticated" event
        """
        if self.clans[sock]["remote_authenticated"] and self.clans[sock]["authenticated"]:
            e = clan_event("host_allowed", host=self.clans[sock]["hostname"])
            self._send_event(e)               

    def handle_auth_reply (self, sock, data) :
        """
        This is the reply after attempting to authenticate to another host
        """
        log("Lets see if i am remote-authenticated: %s " % data.get("authenticated"))
        if data.get("authenticated") == "True" :
            self.clans[sock]["remote_authenticated"] = True
            
        else :
            self.clans[sock]["remote_authenticated"] = False
            e = clan_event("host_denied", host=self.clans[sock]["hostname"])
            self._send_event(e)               
        self.signal_auth(sock)

    def handle_auth_req (self, sock, data) :
        """
        Somebody is sending us an authentication request
        """
        log("Validation for host %s" % data.get("host"))
        host = data.get("host")
        key = data.get("key")
        if self.clans[sock]["hostname"] == "" :
            self.clans[sock]["hostname"] = host
        if self.check_key(sock, host, key) and self.clans[sock]["hostname"] == host :
            reply = self.create_rpc_str("auth_reply", "", {"authenticated":"True"})
            self.clans[sock]["authenticated"] = True
            log("%s is now set to authenticated" % self.clans[sock]["hostname"])               
            self.signal_auth(sock) 
        else :
            reply = self.create_rpc_str("auth_reply", "", {"authenticated":"False"})
            self.clans[sock]["authenticated"] = False
            log("%s has been denied access" % self.clans[sock]["hostname"])
            e = clan_event("host_denied", host=host)
            self._send_event(e)               

        self.queue_req(sock, reply)
        loc_auth = self.clans[sock]["authenticated"]
        rem_auth = self.clans[sock]["remote_authenticated"]
        rem_fail = self.clans[sock]["remote_auth_failed"]
        if loc_auth == True and rem_auth == False and rem_fail == False :
            self.send_auth(host)

    def get_remote_key (self, host): 
        h = self.hosts_conf.hosts
        if not h.get(host):
            log("get_remote_key: %s does not exist in config" % host)
            return ""
        return h[host].get("host_remote_pword")
    

    def check_key(self, sock, host, key) :
        h = self.hosts_conf.hosts
        if not h.get(host):
            log("check_key: %s does not exist in config" % host)
            return ""
        if h[host].get("host_ip") != "" and h[host].get("host_ip") != self.clans[sock]["ip_address"]:
            log("check_key: wrong ip address for %s" % host)
            return False 
        if h[host].get("host_pword") == key:
            return True
        return False

    def remove_rfd (self, sock, rfd) :
        """
        Remove "request for data" entry
        Should be called after an rpc request has finished 
        """
        i = 0
        try :
            for r in self.clans[sock]["rfd"] :
                if r == rfd :
                    self.clans[sock]["rfd"].pop(i)
                i = i + 1
        except :
            log("ERROR in remove_rfd: %s" % traceback.print_exc())         
            pass

    def send_to_gateway (self, res_str) :
        try :
            if not self.config.gateway :
                return False
            gw = self.get_sock_by_hostname(self.config.gateway)
            self.queue_req(gw, res_str)
        except :
            log("Could not forward request to gateway")
            return False
        return True 

    def handle_rpc_req (self, sock, data) :
        """
        Handle incoming rpc request
        """
        if not self.is_authenticated (sock) :
            return "" # return error string !   
        conn = data.get("_conn")
        host = data.get("_h")  # target host (endpoint) of rpc
        call = data.get("_c")
        rfd  = data.get("_rfd")  # request for data
        route_list = data.get("_rl")
        
        if not host and self.is_local(sock):
            host = self.hostname
        #if not self.is_local(sock) and host == self.hostname:
        #    log("seems like %s is pretending to be %s (me)" % (self.clans[sock].get("ip_address"), host))
        #    return
        if self.is_local(sock) and not rfd :
            return "" # return error string !
        else :
            self.clans[sock]["rfd"].append(rfd)
        try :
            data.pop("_c")
            data.pop("_t")
        except :
            pass


        #ra = self.parse_route(route_list)
        # Make sure last sender is last in route list
        #if not self.is_local(sock) and ra[-1] != self.get_hostname_by_sock(sock):
        #    log("%s not last in route list" % ra[-1])
        #    return ""

        if host == self.hostname :
            if not route_list : route_list = ""
            data["_rl"] = "%s+%s>%s" % (route_list,self.get_hostname_by_sock(sock),self.hostname)
            if conn :
                if conn == "close" :
                    log("Got a close request!")
                    #self.send_req(sock, "")
                    res_str = self.create_rpc_resultstr([], rfd)
                    self.queue_req(sock, res_str)
                    self.remove_connection(sock)
                    return ""

            # call local functions
            #r = rpcClan()
            log("%s calls %s" % (self.get_hostname_by_sock(sock), call))
            #r.parent = self # leave it here for backward compatibility
            #r.clan = self 
            res = self.run_rpc(call, data) 

            res_str = self.create_rpc_resultstr(res, rfd)
            #self.send_req(sock, res_str)
            self.queue_req(sock, res_str, rfd)
            #self.remove_rfd(sock, rfd)
        else : # call is not meant for this clan, try and forward it
            if not route_list : route_list = ""
            data["_rl"] = "%s+%s>%s>%s" % (route_list,self.get_hostname_by_sock(sock),self.hostname, host)
            r_sock = self.get_sock_by_hostname(host)
            if not r_sock :
                if self.config.gateway != host :
                    req_str = self.create_rpc_str("rpc", call, data)
                    if self.send_to_gateway(req_str) :
                        log("Request sent to gateway %s for %s" % (self.config.gateway, host))
                        return ""    
                err = self.create_rpc_connerr("Clan is down", rfd) 
                self.queue_req(sock, err)
                log("Could not send data to remote host, clan entry is missing")
                return "" 
            elif not self.is_authenticated(r_sock) :
                err = self.create_rpc_connerr("Clan is not yet authenticated", rfd) 
                self.queue_req(sock, err)
                log("clan not authed yet")
                return "" 
            fwd_host = self.get_hostname_by_sock(sock) 
            if fwd_host == "" :
               fwd_host = "localhost"
            log("Forwarding %s from %s to %s" % (call, fwd_host, host))
            req_str = self.create_rpc_str("rpc", call, data)
            self.queue_req(r_sock, req_str)

    def handle_rpc_reply (self, sock, data, brdata) :
        """
        A host is replying to our rpc call
        """
        if not self.is_authenticated (sock) :
            return # return error string !   
        rfd  = data.get("_rfd")  # request for data
        if not rfd :
            return # return error string !
        r_sock = self.get_sock_by_rfd(rfd)
        if not r_sock :
            log("could not get socket from rfd %s" % rfd)
            return # return error string !
        arr = self.result_to_array(brdata.get_all())
        res_str = self.create_rpc_resultstr(arr, "")
        self.queue_req(r_sock, res_str)

    def result_to_array (self, res) :
        arr = []
        for idx in res :
            if len(res[idx]) != 0 :
                arr.append(res[idx])
        return arr


    def run_rpc (self, method, data) :
        """
        import and execute module/method referenced in rpc call
        """ 
        method = method.replace(" ", "")
        (mod, func) = method.split("::")
        if func[:1] == "_" :
            return self.build_rpc_fault("4", "%s.%s is a private method" % (mod, func))
        if not os.path.isfile("modules/" + mod + ".py") and not os.path.isfile("modules/" + mod + ".pyc") :                                                 
            return self.build_rpc_fault("1", "Module does not exist")
        if not self.validate(mod, func, data):
            return self.build_rpc_fault("1", "Permission denied")

        try :
            if self.config.is_dev: 
                if sys.modules.get(mod) :
                    sys.modules.pop(mod)
                    # sometimes time gets lost
                    import time
            mod_obj = __import__(mod)
            m = eval("mod_obj." + mod + "()")
            try :
                m.parent = self 
                m.clan = self
            except :
                pass                 
            res_args = eval("m." + func + "(data)")
        except :
            return self.build_rpc_fault("3", "Error calling %s.%s: %s" % (mod, func, traceback.print_exc()))

        return res_args

    def parse_route(self, str_list):
        route_array = []
        if not str_list:
            return route_array
        for part in str_list.split("+"):
            for h in part.split(">"):
                if h == "" : continue
                route_array.append(h)
        return route_array                                                                                                                                  
       

    def validate(self, mod, func, data):
        h = self.hosts_conf.hosts
        route_list = data.get("_rl")
        if not route_list :
            return False
        ra = self.parse_route(route_list)
        if len(ra) == 0:
            return False
        orig_host = ra[0]
        if orig_host == self.hostname: # assume this is localhost to localhost
            return True 
        pm = self.permissions
        # this still needs some work, needs to be more solid, test entire route. 

        if not h.has_key(orig_host):
            return False
        if not h[orig_host].has_key("groups"):
            return False
        
        for grp in h[orig_host]["groups"]:
            try:
                pm[grp]["%s::%s" % (mod, func)]
                return True
            except:
                continue
                #return False
        log("%s does not have permission to call %s::%s" % (orig_host, mod, func))
        return False 


    def build_rpc_fault (self, code, text) :
        result = []
        result.append( {"_ERROR":text} )
        return result

    def is_local (self, sock) :
        ip = self.clans[sock]["ip_address"]
        hn = self.clans[sock]["hostname"]
        if ip == "127.0.0.1" and not hn:
            return True
        return False 

    def is_authenticated (self, sock) :
        if self.is_local(sock) :
            return True
        if self.clans[sock]["authenticated"] and self.clans[sock]["remote_authenticated"] :
            return True
        return False   

    def socket_selected (self, sock) :
        #w = False
        for s in self.writers :
            if s == sock :
                return True
        return False

    def queue_req (self, sock, req_str, rfd=None) :
        hostname = self.get_hostname_by_sock(sock)
        #if hostname == "" :
        #    return  
        try:
            self.clans[sock]["send_queue"].append( {"hostname":hostname, "req":req_str, 
                                                 "qtime":self.time(), "rfd":rfd} ) 
        except:
            log(sys.exc_value)

    def send_req (self, sock, msg) :
        """
        Send rpc request to specified socket
        """
        if not self.clans[sock]["connected"] :
            log("Not connected, not sending..")
            return False
        if not self.socket_selected(sock) :
            self.queue_req(sock, msg)
            return False 
        msg = str(len(msg)) + "*" + msg
        #self.readers.remove(sock)
        #self.writers.remove(sock)  
        try :
            sock.sendall(msg)
        except :
            log("Error: (send_req) %s" % traceback.print_exc())
            self.remove_connection(sock)
            return False
        #self.readers.append(sock)
        #self.writers.append(sock)  
        return True

    def get_req (self, sock) :
        """
        Read socket and parse incoming request
        """
        if not self.clans.get(sock) :
            log("seems like conn was removed..")
            return ""
        # to big, break down
        buff = ""
        max_head = 15
        body_len = 0
        head = ""
        try:
            self.readers.remove(sock)
            self.writers.remove(sock)  
        except:
            log("could not remove socket (get_req)")
            return ""
        self.clans[sock]["last_ping"] = self.time() + 5 # dont ping for a while..
        while 1:
            try:		
                buff = sock.recv(1)
            except :
                log("broken connection: %s" % traceback.print_exc())
                self.remove_connection(sock)
                return ""		    
            if buff == "" :
                break
            if buff == "*" :
                body_len = long(head)
                break
            head = head + buff
            if len(head) > max_head :
                break
            continue
        if body_len > 0 :
            bytes_read = long(0)
            buff = ""
            while body_len > bytes_read :
                left = body_len - bytes_read
                try :	
                    
                    tmp_buff = sock.recv(left)


                except :
                    log("broken connection: %s" % traceback.print_exc())
                    self.remove_connection(sock)
                    return ""		    
                bytes_read = bytes_read + len(tmp_buff)
                if bytes_read == 0 :
                    break
                buff = buff + tmp_buff
                
        try:
            self.clans[sock]["last_active"] = self.time()
            self.readers.append(sock)
            self.writers.append(sock)
        except:
            self.remove_connection(sock)

        return buff

    def load_host_config(self):
        self.host_cfg

    def connect_clans (self):
        """
        Connect to each host which has host_direct_connect set to 1
        """
        if not self.hostname: return
        self._reload_configs()
        h = self.hosts_conf.hosts
        for name in h:
            if not h[name].get("host_allowed"):
                continue
            host_ip = h[name].get("host_ip") 
            host_desc = name 
            host_direct_connect = h[name].get("host_direct_connect") 
            host_port = h[name].get("host_port") 
            if not host_desc :
                continue
            sock = self.get_sock_by_hostname(host_desc)
            if sock :
                if self.clans[sock]["connected"] :
                    if (not self.clans[sock]["authenticated"] 
                        or not self.clans[sock]["remote_authenticated"]) and host_direct_connect == 1 :
                        log("resending auth to %s " % host_desc)
                        self.remove_host(host_desc)
                        thread.start_new_thread(self.connect, (host_desc, host_ip, host_port))
                        continue
                    if self.time() - self.clans[sock]["last_active"] > 60:
                        self.remove_host(host_desc)
                        if host_direct_connect == 1:
                            thread.start_new_thread(self.connect, (host_desc, host_ip, host_port))
                            log("reconnecting to %s " % host_desc)
                    elif self.time() - self.clans[sock]["last_ping"] >= 20 :
                        ping = self.create_rpc_str("ping", "", {"time":self.time()})
                        self.queue_req(sock, ping)
                        self.clans[sock]["last_ping"] = self.time()
                    continue
            if host_direct_connect != 1 :
                continue
            if not host_port or not host_ip :
                continue
            #log("going to try and connect to %s" % host_desc)
            thread.start_new_thread(self.connect, (host_desc, host_ip, host_port))
            
    def queue_handler(self, sock):
        s = self.clans[sock]
        while self.clans.has_key(sock):
            if len(s["send_queue"]) == 0:
                time.sleep(0.1)
                continue
            rdict = s["send_queue"].pop(0)  
            hostname = rdict["hostname"]
            req = rdict["req"]
            rfd = rdict["rfd"]
            self.send_req(sock, req)
            #time.sleep(0.1) 
            if rfd and hostname:
                self.remove_rfd(sock, rdict["rfd"])

class clan_event :
    data = None 
    type = None

    def __init__(self, type, **kwargs) :
        self.type = type
        self.data = kwargs

    def get(self, name) :
        return self.data.get(name)
 
    
if __name__ == "__main__" :
        
    import sys
    sys.path.append("modules")
    sys.path.append("http/orders")
    sys.path.append("http")
    sys.path.append("daemons")
    sys.path.append("gui")



    cfg_file = "clanrc"
    cfg = __import__(cfg_file, globals(), locals(), [])
    if not cfg.hostname or not cfg.read_port :
        log("Bad config file, no hostname or read_port values")
    ch = clan_host(cfg.hostname, "",  cfg.read_port)
    ch.connect_clans()
    ch.run()


